package map.sdf2500;

import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import jp.sourceforge.ma38su.util.Log;

import map.City;
import map.Curve;
import map.Road;
import util.JasmineUtil;
import database.FileDatabase;

/**
 * 数値地図2500を読み込むためのクラス
 * @author ma38su
 */
public class Sdf2500Reader {
		
	private static final String KEY_GYOUSEI_ARC = "GYOUSEI_ARC";
	
	private static final String KEY_GYOUSEI_SI_TYO = "GYOUSEI_SI_TYO";
	
	private static final String KEY_GYOUSEI_TYOME = "GYOUSEI_TYOME";
	private static final String KEY_MIZU = "MIZU";
	
	private static final String KEY_MIZU_ARC = "MIZU_ARC";

	private static final String KEY_RAILWAY_ARC = "RAILWAY_ARC";
	private static final String KEY_ROAD_ARC = "ROAD_ARC";

	private static final String KEY_STATION = "STATION_PNT";

	private static final String KEY_TATEMONO = "TATEMONO";

	private static final String KEY_PARK = "ZYOUTI";

	private static final String KEY_RAILBASE = "RAILBASE";
	
	private String charset = "SJIS";

	private FileDatabase db;
	public Sdf2500Reader(FileDatabase db) {
		this.db = db;
	}
	
	/**
	 * アークデータを読み込みます。
	 * @param file ファイル
	 * @return アークデータ
	 * @throws IOException 入出力エラー
	 */
	private Map<Integer, Arc> readArc(File file) throws IOException {
		BufferedReader in = null;
		Map<Integer, Arc> map = new HashMap<Integer, Arc>();
		Pattern csv = Pattern.compile(",");
		try {
			in = new BufferedReader(new FileReader(file));
			String line = in.readLine();
			String[] param = csv.split(line);
//			int flag = Integer.parseInt(param[5]);
			int kei = Integer.parseInt(param[6]);
			double x0 = Double.parseDouble(param[8]); // 8
			double y0 = Double.parseDouble(param[9]); // 9
			int length = 0;
			double[] aryX = null;
			double[] aryY = null;
			int count = 0;
			int code = 0;
			int tag = 0;
			Integer id = null;
			while ((line = in.readLine()) != null) {
				param = csv.split(line);
				if (length == count) {
					if (count > 0 && code != 0) {
						map.put(id, new Arc(aryX, aryY, code, tag));
					}
					count = 0;
					code = Integer.parseInt(param[0].substring(1));
					tag = Integer.parseInt(param[1]); // 線種タグ
					id = Integer.valueOf(param[2]); // 個別番号
					length = Integer.parseInt(param[3]);
					aryX = new double[length];
					aryY = new double[length];
				} else {
					double x = x0 + Double.parseDouble(param[0]);
					double y = y0 + Double.parseDouble(param[1]);
					Point2D point = JasmineUtil.toLatLong(x, y, kei);
					aryX[count] = point.getX();
					aryY[count] = point.getY();
					count++;
				}
			}
			// 最後のひとつを追加
			if (id != null && count == length && count > 0) {
				map.put(id, new Arc(aryX, aryY, code, tag));
			}
		} catch (IOException e) {
			map = null;
			throw e;
		} finally {
			if (in != null) {
				in.close();
			}
		}
		return map;
	}
	
	public ExtendedPolygon[] readTatemono(City city) {
		ExtendedPolygon[] tatemono = null;
		int code = city.getCode();
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_TATEMONO)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_TATEMONO);
			if (obj instanceof Curve[]) {
				tatemono = (ExtendedPolygon[]) obj;
			}
		}
		if (tatemono == null) {
			try {
				File dir = this.db.getSdf2500(code);
				if (dir != null && dir.isDirectory()) {
					Collection<ExtendedPolygon> list = new ArrayList<ExtendedPolygon>();
					for (File file : dir.listFiles()) {
						if (file.isDirectory()) {
							String path = file.getCanonicalPath() + File.separatorChar + "tatemono" + File.separatorChar;
							File dataFile = new File(path + "tatemono.arc");
							if (dataFile.isFile()) {
								Map<Integer, Arc> arcs = this.readArc(dataFile);
								dataFile = new File(path + "tatemono.pgn");
								Log.out(this, "read "+ dataFile);
								Map<Integer, ExtendedPolygon> map = this.readPolygon(dataFile, arcs);
								dataFile = new File(path + "tatemono.atr");
								this.readAttribute(dataFile, map);
								list.addAll(map.values());
							}
						}
					}
					if (list.size() > 0) {
						tatemono = list.toArray(new ExtendedPolygon[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_TATEMONO, tatemono);
					}
				}
			} catch (IOException e) {
				Log.err(this, e);
				tatemono = null;
			}
		}
		if (tatemono != null) {
			city.setTatemono(tatemono);
		}
		return tatemono;
	}
	
	private Map<Integer, ExtendedPoint> readPoint(File file) {
		BufferedReader in = null;
		Map<Integer, ExtendedPoint> map = new HashMap<Integer, ExtendedPoint>();
		try {
			Pattern csv = Pattern.compile(",");
			try {
				in = new BufferedReader(new InputStreamReader(new FileInputStream(file), this.charset));
				String line = in.readLine();
				String[] param = csv.split(line);
				// int flag = Integer.parseInt(param[5]);
				int kei = Integer.parseInt(param[6]);
				double x0 = Double.parseDouble(param[8]); // 8
				double y0 = Double.parseDouble(param[9]); // 9
				while ((line = in.readLine()) != null) {
					param = csv.split(line);
					if (Integer.parseInt(param[0].substring(1)) == 2420) {
						int id = Integer.parseInt(param[1]);
						Point2D p = JasmineUtil.toLatLong(x0 + Double.parseDouble(param[2]), y0 + Double.parseDouble(param[3]), kei);
						map.put(id, new ExtendedPoint((int)(p.getX() * 3600000 + 0.5), (int)(p.getY() * 3600000 + 0.5)));
					}
				}
			} finally {
				if (in != null) {
					in.close();
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		return map;
	}
	private void readAttribute(File file, Map<Integer, ? extends HasAttribute> map) {
		BufferedReader in = null;
		try {
			try {
				in = new BufferedReader(new InputStreamReader(new FileInputStream(file), this.charset));
				Pattern csv = Pattern.compile(",");
				String line = in.readLine();
				while ((line = in.readLine()) != null) {
					String[] param = csv.split(line);
					int id = Integer.valueOf(param[1]);
					HasAttribute arc = map.get(id);
					int size = Integer.parseInt(param[2]);
					if (size > 0) {
						String[] attribute = new String[size];
						for (int i = 0; i < attribute.length && i < param.length - 3; i++) {
							attribute[i] = param[3 + i];
						}
						arc.setAttribute(attribute);
					}
				}
			} finally {
				if (in != null) {
					in.close();
				}
			}
		} catch (IOException e) {
			Log.err(this, e);
		}
	}

	public void readGyousei(City city) {
		Polygon[] si_tyoPolygon = null;
		Polygon[] tyomePolygon = null;
		Curve[] curve = null;
		int code = city.getCode();
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_GYOUSEI_ARC)) {
			if (this.db.hasSerializable(code, Sdf2500Reader.KEY_GYOUSEI_SI_TYO)) {
				si_tyoPolygon = (Polygon[]) this.db.readSerializable(code, Sdf2500Reader.KEY_GYOUSEI_SI_TYO);
			}
			if (this.db.hasSerializable(code, Sdf2500Reader.KEY_GYOUSEI_TYOME)) {
				tyomePolygon = (Polygon[]) this.db.readSerializable(code, Sdf2500Reader.KEY_GYOUSEI_TYOME);
			}
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_GYOUSEI_ARC);
			if (obj instanceof Curve[]) {
				curve = (Curve[]) obj;
			}
		}
		if (si_tyoPolygon == null || tyomePolygon == null) {
			List<Curve> list = new ArrayList<Curve>();
			try {
				File dir = this.db.getSdf2500(code);
				if (dir != null && dir.isDirectory()) {
					Collection<Polygon> tyome = new ArrayList<Polygon>();
					Collection<Polygon> si_tyo = new ArrayList<Polygon>();
					for (File file : dir.listFiles()) {
						if (file.isDirectory()) {
							String path = file.getCanonicalPath() + File.separatorChar + "gyousei" + File.separatorChar;
							File dataFile = new File(path + "gyousei.arc");
							if (dataFile.isFile()) {
								Map<Integer, Arc> arcs = this.readArc(dataFile);
								dataFile = new File(path + "tyome.pgn");
								Log.out(this, "read "+ dataFile);
								Map<Integer, ExtendedPolygon> map = this.readPolygon(dataFile, arcs);
								tyome.addAll(map.values());
								for (Arc arc : arcs.values()) {
									list.add(arc.transformCurve());
								}
								dataFile = new File(path + "si_tyo.pgn");
								Log.out(this, "read "+ dataFile);
								map = this.readPolygon(dataFile, arcs);
								si_tyo.addAll(map.values());
								for (Arc arc : arcs.values()) {
									list.add(arc.transformCurve());
								}
							}
						}
					}
					if (list.size() > 0) {
						curve = list.toArray(new Curve[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_GYOUSEI_ARC, curve);
					}
					if (tyome.size() > 0) {
						tyomePolygon = tyome.toArray(new Polygon[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_GYOUSEI_TYOME, tyomePolygon);
					}
					if (si_tyo.size() > 0) {
						si_tyoPolygon = si_tyo.toArray(new Polygon[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_GYOUSEI_SI_TYO, si_tyoPolygon);
					}
				}
			} catch (IOException e) {
				Log.err(this, e);
				si_tyoPolygon = null;
				tyomePolygon = null;
				curve = null;
			}
		}
		if (tyomePolygon != null && si_tyoPolygon != null && curve != null) {
			city.setGyosei(curve, si_tyoPolygon, tyomePolygon);
		}
	}
	public void readMizu(City city) {
		Polygon[] polygon = null;
		Curve[] curve = null;
		int code = city.getCode();
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_MIZU) && this.db.hasSerializable(code, Sdf2500Reader.KEY_MIZU_ARC)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_MIZU);
			if (obj instanceof Polygon[]) {
				polygon = (Polygon[]) obj;
			}
			obj = this.db.readSerializable(code, Sdf2500Reader.KEY_MIZU_ARC);
			if (obj instanceof Curve[]) {
				curve = (Curve[]) obj;
			}
		}
		if (polygon == null || curve == null) {
			Collection<Curve> mizuArc = new ArrayList<Curve>();
			Collection<Polygon> mizu = new ArrayList<Polygon>();
			try {
				File dir = this.db.getSdf2500(code);
				if (dir != null && dir.isDirectory()) {
					for (File file : dir.listFiles()) {
						if (file.isDirectory()) {
							String path = file.getCanonicalPath() + File.separatorChar + "mizu" + File.separatorChar;
							File dataFile = new File(path + "mizu.arc");
							if (dataFile.isFile()) {
								Map<Integer, Arc> arcs = this.readArc(dataFile);
								Log.out(this, "read "+ dataFile);
								dataFile = new File(path + "mizu.pgn");
								Map<Integer, ExtendedPolygon> map = this.readPolygon(dataFile, arcs);
								mizu.addAll(map.values());
								Log.out(this, "read "+ dataFile);
								for (Arc arc : arcs.values()) {
									mizuArc.add(arc.transformCurve());
								}
							}
						}
					}
					if (mizu.size() > 0) {
						polygon = mizu.toArray(new Polygon[]{});
						curve = mizuArc.toArray(new Curve[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_MIZU, polygon);
						this.db.writeSerializable(code, Sdf2500Reader.KEY_MIZU_ARC, curve);
					}
				}
			} catch (IOException e) {
				Log.err(this, e);
				polygon = null;
				curve = null;
			}
		}
		if (polygon != null && curve != null) {
			city.setMizu(polygon, curve);
		}
	}

	public void readOthers(City city) {
		Railway[][] railways = null;
		ExtendedPoint[] stations = null;
		ExtendedPolygon[] park = null;
		ExtendedPolygon[] railbase = null;
		int code = city.getCode();
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_RAILWAY_ARC)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_RAILWAY_ARC);
			if (obj instanceof Curve[][]) {
				railways = (Railway[][]) obj;
			}
		}
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_STATION)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_STATION);
			if (obj instanceof ExtendedPoint[]) {
				stations = (ExtendedPoint[]) obj;
			}
		}
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_PARK)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_PARK);
			if (obj instanceof ExtendedPolygon[]) {
				park = (ExtendedPolygon[]) obj;
			}
		}
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_RAILBASE)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_RAILBASE);
			if (obj instanceof ExtendedPolygon[]) {
				railbase = (ExtendedPolygon[]) obj;
			}
		}
		if (railways == null || stations == null || park == null || railbase == null) {
			Collection<Curve> other = new ArrayList<Curve>();
			Collection<Curve> jr = new ArrayList<Curve>();
			Collection<Arc> curve = new HashSet<Arc>();
			Collection<ExtendedPoint> station = new ArrayList<ExtendedPoint>();
			Map<Point2D, ArcNode> points = new HashMap<Point2D, ArcNode>();
			Collection<ExtendedPolygon> parkList = new ArrayList<ExtendedPolygon>();
			Collection<ExtendedPolygon> railbaseList = new ArrayList<ExtendedPolygon>();
			try {
				File dir = this.db.getSdf2500(code);
				if (dir != null && dir.isDirectory()) {
					for (File file : dir.listFiles()) {
						// TODO
						Set<String> names = new HashSet<String>();
						if (file.isDirectory()) {
							String path = file.getCanonicalPath() + File.separatorChar + "others" + File.separatorChar;
							File ekiFile = new File(path + "eki.pnt");
							if (ekiFile.isFile()) {
								Map<Integer, ExtendedPoint> map = this.readPoint(ekiFile);
								File atrFile = new File(path + "eki.atr");
								if (atrFile.isFile()) {
									this.readAttribute(atrFile, map);
								}
								station.addAll(map.values());
							}
							File dataFile = new File(path + "others.arc");
							if (dataFile.isFile()) {
								Map<Integer, Arc> arcs = this.readArc(dataFile);
								dataFile= new File(path + "tetudou.atr");
								if (dataFile.isFile()) {
									this.readAttribute(dataFile, arcs);
								}
								dataFile = new File(path + "zyouti.pgn");
								if (dataFile.isFile()) {
									Map<Integer, ExtendedPolygon> map = this.readPolygon(dataFile, arcs);
									dataFile = new File(path + "zyouti.atr");
									this.readAttribute(dataFile, map);
									for (ExtendedPolygon polygon : map.values()) {
										switch (polygon.getCode()) {
											case 6241: // 鉄道敷
												railbaseList.add(polygon);
												break;
											case 6242: // 都市公園
												parkList.add(polygon);
												break;
											case 6243: // 学校敷地
											case 6244: // 神社・寺院境内
											case 6215: // 墓地
											case 6200: // その他の場地
												break;
											default:
												System.out.println(polygon.getCode());
										}
									}
								}
								for (Arc arc : arcs.values()) {
									switch (arc.getCode()) {
									case 2300: // 鉄道
										double[] curveX = arc.getArrayX();
										double[] curveY = arc.getArrayY();
										ArcNode p1 = new ArcNode(curveX[0], curveY[0]);
										ArcNode p2 = new ArcNode(curveX[curveX.length - 1], curveY[curveX.length - 1]);
										if (p1.equals(p2)) {
											curve.add(arc);
										} else {
											if (points.containsKey(p1)) {
												p1 = points.get(p1);
											} else {
												points.put(p1, p1);
											}
											if (points.containsKey(p2)) {
												p2 = points.get(p2);
											} else {
												points.put(p2, p2);
											}
											Arc b = p1.put(p2, arc);
											if (b != null) {
												curve.add(b);
											}
										}
										break;
									case 6241: // 鉄道敷
									case 6242: // 都市公園
									case 6243: // 学校敷地
									case 6244: // 神社・寺院境内
									case 6215: // 墓地
									case 6200: // その他の場地
										break;
									default:
										System.out.println(arc.getCode());
									}
								}
							}
						}
						if (names.size() > 0) {
							Log.out(this, "rail: "+ names);
						}
					}
					this.jointNode(points.values(), curve);
					for (Arc arc : curve) {
						Railway rail = arc.transformRailway();
						if (rail.getType() == 0) {
							jr.add(rail);
						} else {
							other.add(rail);
						}
					}
					if (other.size() + jr.size() > 0) {
						railways = new Railway[][]{jr.toArray(new Railway[]{}), other.toArray(new Railway[]{})};
						this.db.writeSerializable(code, Sdf2500Reader.KEY_RAILWAY_ARC, railways);
					}
					if (station.size() > 0) {
						stations = station.toArray(new ExtendedPoint[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_STATION, stations);
					}
					if (parkList.size() > 0) {
						park = parkList.toArray(new ExtendedPolygon[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_PARK, park);
					}
					if (railbaseList.size() > 0) {
						railbase = railbaseList.toArray(new ExtendedPolygon[]{});
						this.db.writeSerializable(code, Sdf2500Reader.KEY_RAILBASE, railbase);
					}
				}
			} catch (IOException e) {
				Log.err(this, e);
				railways = null;
			}
		}
		if (railways != null) {
			city.setRailway(railways);
		}
		if (stations != null) {
			city.setStation(stations);
		}
		if (park != null) {
			city.setPark(park);
		}
		if (railbase != null) {
			city.setRailbase(railbase);
		}
	}
	
	/**
	 * グラフをつなぎあわせて頂点数を減らします。
	 * 次数2の頂点は削除して、次数1と、次数3の頂点のみ残します。
	 * @param nodes つなぎあわせる頂点
	 * @param set
	 */
	private void jointNode(Collection<ArcNode> nodes, Collection<Arc> set) {
		Set<Arc> borderSet = new HashSet<Arc>();
		for (ArcNode node : nodes) {
			if (node.getBorder().size() == 2) {
				for (int i = 0; i < 4; i++) {
					Arc border = node.connect(i);
					if (border != null) {
						set.add(border);
					}
				}
			}
		}
		for (ArcNode node : nodes) {
			for (Arc border : node.getBorder()) {
				borderSet.add(border);
			}
		}
		for (Arc border : borderSet) {
			set.add(border);
		}
	}


	/**
	 * ポリゴンデータを読み込みます。
	 * @param polygon 読み込むポリゴンデータ
	 * @param file 
	 * @param arcs アークデータ
	 * @return ポリゴンデータ
	 * @throws IOException 入出力エラー
	 */
	private Map<Integer, ExtendedPolygon> readPolygon(File file, Map<Integer, Arc> arcs) throws IOException {
		Map<Integer, ExtendedPolygon> map = new HashMap<Integer, ExtendedPolygon>();
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(file), this.charset));
			// ヘッダを読み飛ばす
			bi.readLine();
			int length = 0;
			int count = 0;
			Integer id = null;
			Arc[] aryArc = null;
			Pattern csv = Pattern.compile(",");
			int code = 0;
			while (bi.ready()) {
				String line = bi.readLine();
				if (length == count) {
					if (count > 0 && id != null) {
						map.put(id, Arc.transformPolygon(code, aryArc));
					}
					String[] param = csv.split(line);
					count = 0;
					code = Integer.parseInt(param[0].substring(1));
					id = Integer.valueOf(param[1]);
					length = Integer.parseInt(param[2]);
					aryArc = new Arc[length];
				} else {
					int key = Integer.parseInt(line);
					Arc arc;
					if (key < 0) {
						arc = arcs.get(- key).reverse();
					} else {
						arc = arcs.get(key);
					}
					if (arc == null) {
						throw new IllegalArgumentException("エラー");
					}
					aryArc[count] = arc;
					count++;
				}
			}
			if (count > 0 && length == count && id != null) {
				map.put(id, Arc.transformPolygon(code, aryArc));
			}
		} catch (IOException e) {
			map.clear();
			throw e;
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return map;
	}
	
	public void readRoad(City city) {
		Road[][] roads = null;
		int code = city.getCode();
		if (this.db.hasSerializable(code, Sdf2500Reader.KEY_ROAD_ARC)) {
			Object obj = this.db.readSerializable(code, Sdf2500Reader.KEY_ROAD_ARC);
			if (obj instanceof Road[][]) {
				roads = (Road[][]) obj;
			}
		}
		if (roads == null) {
			int maxWidth = 4;
			List<List<Curve>> normal = new ArrayList<List<Curve>>(maxWidth);
			List<List<Curve>> main = new ArrayList<List<Curve>>(maxWidth);
			List<List<Curve>> highway = new ArrayList<List<Curve>>(maxWidth);
			for (int i = 0; i < maxWidth; i++) {
				normal.add(new ArrayList<Curve>());
				main.add(new ArrayList<Curve>());
				highway.add(new ArrayList<Curve>());
			}
			try {
				File dir = this.db.getSdf2500(code);
				if (dir != null && dir.isDirectory()) {
					for (File file : dir.listFiles()) {
						if (file.isDirectory()) {
							String path = file.getCanonicalPath() + File.separatorChar + "road" + File.separatorChar;
							File arcFile = new File(path + "roadntwk.arc");
							if (arcFile.isFile()) {
								Map<Integer, Arc> arcs = this.readArc(arcFile);
								File atrFile = new File(path + "road.atr");
								if (atrFile.isFile()) {
									this.readAttribute(atrFile, arcs);
								}
								for (Arc arc : arcs.values()) {
									Road road = arc.transformRoad();
									if (road.getType() == 4) {
										highway.get(road.getWidth() - 1).add(road);
									} else if (road.getType() == 3) {
										main.get(road.getWidth() - 1).add(road);
									} else {
										normal.get(road.getWidth() - 1).add(road);
									}
								}
							}
						}
					}
					if ((highway.size() + main.size() + normal.size()) > 0) {
						roads = new Road[maxWidth * 3][];
						int index = 0;
						for (int i = 0; i < normal.size(); i++) {
							roads[i] = normal.get(i).toArray(new Road[]{});
						}
						index += normal.size();
						for (int i = 0; i < main.size(); i++) {
							roads[i + index] = main.get(i).toArray(new Road[]{});
						}
						index += main.size();
						for (int i = 0; i < highway.size(); i++) {
							roads[i + index] = highway.get(i).toArray(new Road[]{});
						}
						this.db.writeSerializable(code, Sdf2500Reader.KEY_ROAD_ARC, roads);
					}
				}
			} catch (IOException e) {
				Log.err(this, e);
				roads = null;
			}
		}
		if (roads != null) {
			city.setRoad(roads);
		}
	}
}
